/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "soustirage.h"

//fonction renvoyant la taille d'un tirage
int taille(const tirage &t){
	int resultat=0;
	tirage::const_iterator it;
	
	for(it=t.begin();it!=t.end();it++){
		resultat=resultat+(it->second);
	}
	
	return resultat;
}

//fonction prenant un tirage en argument, et renvoyant un vecteur
//contenant les sous-tirages de ce tirage, tris dans un ordre "intressant"
vector<tirage> cree_vect(const tirage &t){
	tirage tvide;
	
	//cre un vecteur contenant uniquement le tirage vide
	vector<tirage> v(1,tvide);
	
	tirage::const_iterator i;
	for(i=t.begin();i!=t.end();i++){	//boucles traduisant un algorithme mathmatique qu'un papier aide  comprendre
						//mfions-nous, la taille de v varie tout au long de la boucle indexe par j
		int a=v.size();
		int j; int k;
		for(j=1;j<=i->second;j++){
			for(k=0;k<a;k++){
				tirage temp=v[k];
				temp[i->first]=j;
				v.push_back(temp);
			}
		}
	}
	
	return v;
}

//fonction affichant un tirage
ostream& operator<<(ostream& ostr, const tirage& t) {
	
	if (t.size()==0) {
		return ostr << "vide";
	}
	
	tirage::const_iterator i; int j;
	for(i=t.begin(); i!=t.end(); i++){
		for(j=0; j<(i->second); j++){
			ostr << (int) (i->first) << ".";
		}
	}
	
	return ostr << "\b";
}

//fonction affichant un vecteur de tirages.
ostream& operator<<(ostream& ostr, const vector<tirage>& v){
	int taille=v.size(); int i;
	
	ostr << "[";
	for(i=0; i<taille-1; i++) {
		ostr << v[i] << " ; ";
	}
	ostr << v[taille-1];
	ostr << "]";
	
	return ostr;
}

//renvoie true si le tirage est valable (pas trop de plaques de chaque sorte).
bool est_valable(tirage t){
	bool resultat=true;
	int i;
	for (i=1;i<=10;i++) { if(t[i]>2) {resultat=false;} }
	if (t[25]>1) {resultat=false;}
	if (t[50]>1) {resultat=false;}
	if (t[75]>1) {resultat=false;}
	if (t[100]>1) {resultat=false;}
	return resultat;
}

//cree les 134596 faons de choisir 6 nb parmi 0;1;2;...;23
list<pseudo_tirage> cree_liste_pseudo(){
	
	//initialisation de v
	pseudo_tirage v;
	for(int i=0;i<6;i++){
		v.push_back(i);
	}
	
	//initialisation de la liste
	list<pseudo_tirage> liste;
	int pos=5;
	bool pas_termine=true;
	
	while(pas_termine){
		while(v[5]<24){
			liste.push_back(v);
			v[5]++;
		}
	
		while(pas_termine && v[pos]==19+pos){
			pos--;
			if(pos>=0){
				v[pos]++;
			}
			else{
				pas_termine=false; pos++;
			}
		}
		
		if(pas_termine){
			for(int j=pos+1;j<6;j++){
				v[j]=v[pos]+j-pos;
			}
		}
		
		pos=5;
	}
	
	//cout<<"nb de listes trouves : " << liste.size() << endl;
	return liste;
}

//transforme un pseudo-tirage en tirage
tirage pseudo_en_tirage(pseudo_tirage p){
	
	tirage t; int i;
	
	for(i=0;i<6;i++){
		switch (p[i]){
			case 23: t[100]=1; break;
			case 22: t[75]=1; break;
			case 21: t[50]=1; break;
			case 20: t[25]=1; break;
			default: t[1+p[i]/2]++; break;
		}
	}
	
	return t;
}

//stocke les 13243 tirages diffrents
set<tirage> tous_les_tirages(){
	
	set<tirage> resultat;
	
	list<pseudo_tirage> liste_pseudo=cree_liste_pseudo();
	list<pseudo_tirage>::iterator PT;
	
	for(PT=liste_pseudo.begin();PT!=liste_pseudo.end();PT++){
		resultat.insert(pseudo_en_tirage(*PT));
	}
	
	//cout << resultat.size() << endl;
	
	return resultat;
}

//renvoie le poids d'un tirage relativement au nb de couples dans ce tirage,
//en supposant que ce tirage est un tirage possible dans le jeu (pas de triplettes ou plus)
int coefficient(tirage t){
	int resultat=1;
	
	tirage::iterator it;
	
	for(it=t.begin();it!=t.end();it++){
		if(it->second==1 && it->first<11){
			resultat=resultat*2;
		}
	}
	
	return resultat;
}

//fonction renvoyant la plus petite plaque suprieure  une plaque p, et disponible une fois enleves les plaques du tirage t
plaque plus_petite_plaque_sup(tirage t,plaque p) {
	plaque resultat=augmente_plaque(p);
	while ( (t[resultat]==2 && resultat<=10) || (t[resultat]==1 && resultat >10) ){
		resultat=augmente_plaque(resultat);
	}
	if(resultat<p) resultat=0;  //commenter si on autorise le bouclage
	return resultat;
}

//fonction renvoyant la plus petite plaque infrieure  une plaque p, et disponible une fois enleves les plaques du tirage t
plaque plus_petite_plaque_inf(tirage t,plaque p) {
	plaque resultat=diminue_plaque(p);
	while ( (t[resultat]==2 && resultat<=10) || (t[resultat]==1 && resultat >10) ){
		resultat=diminue_plaque(resultat);
	}
	if(resultat>p) resultat=0;  //commenter si on autorise le bouclage
	return resultat;
}

//fonction donnant la plaque juste plus grande que la plaque p (avec bouclage)
plaque augmente_plaque(plaque p) {
	switch(p) {
		case  10: return 25;
		case  25: return 50;
		case  50: return 75;
		case  75: return 100;
		case 100: return 1;
		default : return p+1;
	}
}

//fonction donnant la plaque juste plus petite que la plaque p (avec bouclage)
plaque diminue_plaque(plaque p) {
	switch(p) {
		case  25: return 10;
		case  50: return 25;
		case  75: return 50;
		case 100: return 75;
		case   1: return 100;
		default : return p-1;
	}
}
